#!/usr/bin/python
# Nom de fichier : bwa-multi_wrapper.py

"""
Runs BWA on single-end or paired-end data.
Produces either one BAM file containing all the mappings or one BAM for each mapping
Works with BWA version 0.5.9.

usage: bwa_wrapper.py [options]

See below for options
"""

from time import localtime, strftime
import optparse, os, shutil, subprocess, sys, tempfile


def stop_err( msg ):
	sys.stderr.write( '%s\n' % msg )
	sys.exit()

def check_is_double_encoded( fastq ):
	# check that first read is bases, not one base followed by numbers
	bases = [ 'A', 'C', 'G', 'T', 'a', 'c', 'g', 't', 'N' ]
	nums = [ '0', '1', '2', '3' ]
	for line in file( fastq, 'rb'):
		if not line.strip() or line.startswith( '@' ):
			continue
		if len( [ b for b in line.strip() if b in nums ] ) > 0:
			return False
		elif line.strip()[0] in bases and len( [ b for b in line.strip() if b in bases ] ) == len( line.strip() ):
			return True
		else:
			raise Exception, 'First line in first read does not appear to be a valid FASTQ read in either base-space or color-space'
	raise Exception, 'There is no non-comment and non-blank line in your FASTQ file'

def getListFromFile (list_of_reads_file, pairing, forwardReadsFileList, reverseReadsFileList,IndivNames  ):
	# This function reads an input file that contains the complete path of a reads file on each line if the data is single-end .
	# When data are paired each line contains a pair of tab separated dataset names (one for forward and one for the revese file)
	# One or two lists of reads file are built during the process
	
	# Initializations
	reads_list = []
	
	# Build list
	try:
		read_file_obj = open(list_of_reads_file, 'r')
	except:
		print 'Error: Cannot open/read file: ', list_of_reads_file
	
	grp = 0
	for line in read_file_obj.readlines():
		fastafilename = line.strip()
		if fastafilename != "":
			grp = grp + 1
			if pairing != "paired":
				forwardReadsFileList.append(fastafilename)
			else:
				dataliste = fastafilename.split("\t")
				forwardReadsFileList.append(dataliste[0])
				reverseReadsFileList.append(dataliste[1])
				if (len(dataliste) > 2):
					IndivNames.append(dataliste[2].strip())
				else:
					IndivNames.append(str(grp) )

						
	read_file_obj.close()
	
	return 

def runCmd(cmd, tmp_dir):
	print (cmd )	
	tmp = tempfile.NamedTemporaryFile( dir=tmp_dir ).name
	tmp_stderr = open( tmp, 'wb' )
	
	proc = subprocess.Popen( args=cmd, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno() )
	returncode = proc.wait()
	if returncode != 0:
			print ("  --> A fatal error occured during the execution procedure..\n")
			raise Exception, stderr
	else:
			print ("  --> Execution complete !\n")
	tmp_stderr.close()	
	return 

def __main__():
	
	# Initializations
	version = "2.0"
	
	# Welcome message
	print("################################################################")
	print("#          BWA-Multi Galaxy Wrapper  (Version " + version + ")         #")
	print("################################################################")
	print "Start time: ", strftime("%d-%m-%Y_%H:%M:%S", localtime()), "\n"
	
	#Parse Command Line
	parser = optparse.OptionParser()
	parser.add_option( '-t', '--threads', dest='threads', help='The number of threads to use' )
	parser.add_option( '-c', '--color-space', dest='color_space', action='store_true', help='If the input files are SOLiD format' )
	parser.add_option( '-r', '--ref', dest='ref', help='The reference genome to use or index' )
	parser.add_option( '-f', '--input', dest='fastq', help='The (forward) fastq file to use for the mapping' )
	
	parser.add_option( '-u', '--output', dest='output', help='The file to save the output (BAM format)' )
	parser.add_option( '', '--output_id', dest='outputID', help='The ID of file to save the output (BAM format)' )
	parser.add_option( '', '--database_tmp_dir', dest='outputDIR', help='The Tool Galaxy dir' )
	parser.add_option( '', '--output_grp', dest='outputGRP', help='The list of read groups in the bam' )
	parser.add_option( '', '--mergebams', dest='mergeBAM', help='On or multiple bam' )
    
	parser.add_option( '-g', '--genAlignType', dest='genAlignType', help='The type of pairing (single or paired)' )
	parser.add_option( '-p', '--params', dest='params', help='Parameter setting to use (pre_set or full)' )
	parser.add_option( '-s', '--fileSource', dest='fileSource', help='Whether to use a previously indexed reference sequence or one form history (indexed or history)' )
	parser.add_option( '-n', '--maxEditDist', dest='maxEditDist', help='Maximum edit distance if integer' )
	parser.add_option( '-m', '--fracMissingAligns', dest='fracMissingAligns', help='Fraction of missing alignments given 2% uniform base error rate if fraction' )
	parser.add_option( '-o', '--maxGapOpens', dest='maxGapOpens', help='Maximum number of gap opens' )
	parser.add_option( '-e', '--maxGapExtens', dest='maxGapExtens', help='Maximum number of gap extensions' )
	parser.add_option( '-d', '--disallowLongDel', dest='disallowLongDel', help='Disallow a long deletion within specified bps' )
	parser.add_option( '-i', '--disallowIndel', dest='disallowIndel', help='Disallow indel within specified bps' )
	parser.add_option( '-l', '--seed', dest='seed', help='Take the first specified subsequences' )
	parser.add_option( '-k', '--maxEditDistSeed', dest='maxEditDistSeed', help='Maximum edit distance to the seed' )
	parser.add_option( '-M', '--mismatchPenalty', dest='mismatchPenalty', help='Mismatch penalty' )
	parser.add_option( '-O', '--gapOpenPenalty', dest='gapOpenPenalty', help='Gap open penalty' )
	parser.add_option( '-E', '--gapExtensPenalty', dest='gapExtensPenalty', help='Gap extension penalty' )
	parser.add_option( '-R', '--suboptAlign', dest='suboptAlign', help='Proceed with suboptimal alignments even if the top hit is a repeat' )
	parser.add_option( '-N', '--noIterSearch', dest='noIterSearch', help='Disable iterative search' )
	parser.add_option( '-T', '--outputTopN', dest='outputTopN', help='Maximum number of alignments to output in the XA tag for reads paired properly' )
	parser.add_option( '', '--outputTopNDisc', dest='outputTopNDisc', help='Maximum number of alignments to output in the XA tag for disconcordant read pairs (excluding singletons)' )
	parser.add_option( '-S', '--maxInsertSize', dest='maxInsertSize', help='Maximum insert size for a read pair to be considered mapped good' )
	parser.add_option( '-P', '--maxOccurPairing', dest='maxOccurPairing', help='Maximum occurrences of a read for pairings' )
	parser.add_option( '-D', '--dbkey', dest='dbkey', help='Dbkey for reference genome' )
	parser.add_option( '-X', '--do_not_build_index', dest='do_not_build_index', action='store_true', help="Don't build index" )
	parser.add_option( '-U', '--do_remove_unmapped', dest='do_remove_unmapped',  help="Don't output unmapped reads" )
	

	(options, args) = parser.parse_args()
	
	# Get the list of reads file (forward and / or reverse)
	forwardReadsFileList = []
	reverseReadsFileList = []
	IndivNames           = []
	
	if options.fastq:
		getListFromFile(options.fastq, options.genAlignType, forwardReadsFileList, reverseReadsFileList,IndivNames  )
		

	# output version # of tool
	try:
		tmp = tempfile.NamedTemporaryFile(dir="./").name
		tmp_stdout = open( tmp, 'wb' )
		proc = subprocess.Popen( args='bwa 2>&1', shell=True, stdout=tmp_stdout )
		tmp_stdout.close()
		returncode = proc.wait()
		stdout = None
		for line in open( tmp_stdout.name, 'rb' ):
			if line.lower().find( 'version' ) >= 0:
				stdout = line.strip()
				break
		if stdout:
			print('BWA ' + stdout + "\n")
		else:
			raise Exception
	except:
		sys.stdout.write( 'Could not determine BWA version\n' )
		
	# Indicate if BWA run on single-end or paired-end
	print ('BWA run on ' + options.genAlignType + '-end data' + "\n")

	# check for color space fastq that's not double-encoded and exit if appropriate
	if options.color_space:
		for fastqfile in forwardReadsFileList:
			if not check_is_double_encoded( fastqfile ):
				stop_err( fastqfile )
		if options.genAlignType == 'paired':
			for fastqfile in reverseReadsFileList:
				if not check_is_double_encoded( fastqfile ):
					stop_err( 'Your reverse reads file must also be double-encoded (it must be converted from "numbers" to "bases"). See the help section for details' )


	# set color space variable
	if options.color_space:
		color_space = '-c'
	else:
		color_space = ''

	# make temp directory for placement of indices
	tmp_index_dir = tempfile.mkdtemp(dir="./")
	tmp_dir = tempfile.mkdtemp("./")
	
	# index if necessary
	if options.fileSource == 'history' and not options.do_not_build_index:
		print('Fasta file indexation procedure')
		ref_file = tempfile.NamedTemporaryFile( dir=tmp_index_dir )
		ref_file_name = ref_file.name
		ref_file.close()
		os.symlink( options.ref, ref_file_name )
		# determine which indexing algorithm to use, based on size
		try:
			size = os.stat( options.ref ).st_size
			if size <= 2**30: 
				indexingAlg = 'is'
			else:
				indexingAlg = 'bwtsw'
		except:
			indexingAlg = 'is'
		indexing_cmds = '%s -a %s' % ( color_space, indexingAlg )
		cmd1 = 'bwa index %s %s' % ( indexing_cmds, ref_file_name )
		print (' -> BWA index will be executed with the following command line: ' + cmd1)

		try:
			tmp = tempfile.NamedTemporaryFile( dir=tmp_index_dir ).name
			
			# Execute BWA index
			tmp_stderr = open( tmp, 'wb' )
			proc = subprocess.Popen( args=cmd1, shell=True, cwd=tmp_index_dir, stderr=tmp_stderr.fileno() )
			returncode = proc.wait()
			if returncode != 0:
				print ("  --> A fatal error occured during the execution procedure..\n")
				raise Exception, stderr
			else:
				print ("  --> Execution complete !\n")
			tmp_stderr.close()
			
			# get stderr, allowing for case where it's very large
			tmp_stderr = open( tmp, 'rb' )
			stderr = ''
			buffsize = 1048576
			try:
				while True:
					stderr += tmp_stderr.read( buffsize )
					if not stderr or len( stderr ) % buffsize != 0:
						break
			except OverflowError:
				pass
			tmp_stderr.close()

			
		except Exception, e:
			# clean up temp dirs
			if os.path.exists( tmp_index_dir ):
				shutil.rmtree( tmp_index_dir )
			if os.path.exists( tmp_dir ):
				shutil.rmtree( tmp_dir )
			stop_err( 'Error indexing reference sequence. ' + str( e ) )
	else:
		ref_file_name = options.ref

	# set up aligning and generate aligning command options
	if options.params == 'pre_set':
		aligning_cmds = '-t %s %s' % ( options.threads, color_space )
		gen_alignment_cmds = ''
	else:
		if options.maxEditDist != '0':
			editDist = options.maxEditDist
		else:
			editDist = options.fracMissingAligns
		if options.seed != '-1':
			seed = '-l %s' % options.seed
		else:
			seed = ''
		if options.suboptAlign == 'true':
			suboptAlign = '-R'
		else:
			suboptAlign = ''
		if options.noIterSearch == 'true':
			noIterSearch = '-N'
		else:
			noIterSearch = ''
		aligning_cmds = '-n %s -o %s -e %s -d %s -i %s %s -k %s -t %s -M %s -O %s -E %s %s %s %s' % \
						( editDist, options.maxGapOpens, options.maxGapExtens, options.disallowLongDel,
							options.disallowIndel, seed, options.maxEditDistSeed, options.threads,
							options.mismatchPenalty, options.gapOpenPenalty, options.gapExtensPenalty,
							suboptAlign, noIterSearch, color_space )
		if options.genAlignType == 'paired':
			gen_alignment_cmds = '-a %s -o %s' % ( options.maxInsertSize, options.maxOccurPairing )
			if options.outputTopNDisc:
				gen_alignment_cmds += ' -N %s' % options.outputTopNDisc
		else:
			gen_alignment_cmds = ''
			
		if options.outputTopN:
			gen_alignment_cmds += ' -n %s' % options.outputTopN
	
		
		
	# prepare actual aligning and generate aligning commands
	print("Command line building (BWA aln forward/reverse and samse/sampe)\n")
	
	cmd2 = [] # BWA aln forward
	cmd2b = [] # BWA aln reverse
	cmd3 = [] # BWA samse/sampe
	result = []
	
	grp = 1
	if options.mergeBAM == "yes":
		rgfile = open(options.outputGRP, "w")
		
	if options.genAlignType != 'paired':
		for forfastqfile in forwardReadsFileList:
			# set up output files
			tmp_align_out = tempfile.NamedTemporaryFile( dir=tmp_dir )
			tmp_align_out_name=tmp_align_out.name
			tmp_align_out.close()
			
			tmp_out = tempfile.NamedTemporaryFile( dir=tmp_dir )
			tmp_out_name=tmp_out.name
			tmp_out.close()
			
			datasetname   = os.path.splitext(os.path.basename(fastqfile))[0]
			datasetname   = datasetname.replace('_', '-') + '-' + str(grp)
			if options.mergeBAM != "yes":
				bam_file_name =  'primary_%s_%s_visible_bam' % ( options.outputID, datasetname)
				bam_out_name  =  os.path.join( options.outputDIR, bam_file_name)
			else:
				bam_file_name = datasetname
				bam_out_name  = os.path.join( tmp_dir, bam_file_name)
				readGroup = '@RG\tID:%s\tSM:%s' % ( bam_file_name, datasetname )
				rgfile.write(readGroup+"\n")

			
			result.append(bam_out_name)
			
			command = gen_alignment_cmds 
			
			cmd2.append( 'bwa aln %s %s %s > %s' % ( aligning_cmds, ref_file_name, fastqfile, tmp_align_out_name ) )
			
			if options.do_remove_unmapped=="yes":
				cmd3.append( 'bwa samse %s %s %s %s | samtools view -bSh -F 4 - >> %s' % ( command, ref_file_name, tmp_align_out_name, fastqfile, bam_out_name ) )
			else:
				cmd3.append( 'bwa samse %s %s %s %s | samtools view -bSh - >> %s' % ( command, ref_file_name, tmp_align_out_name, fastqfile, bam_out_name ) )
			grp = grp + 1 ;
	else:	
		for forfastqfile, revfastqfile,IndivName in zip(forwardReadsFileList, reverseReadsFileList,IndivNames):
			# set up output files
			
			tmp_align_out      = tempfile.NamedTemporaryFile( dir=tmp_dir )
			tmp_align_out_name = tmp_align_out.name
			tmp_align_out.close()
			
			tmp_out      = tempfile.NamedTemporaryFile( dir=tmp_dir )
			tmp_out_name = tmp_out.name
			tmp_out.close()
			
			datasetname   = os.path.splitext(os.path.basename(forfastqfile))[0]+'-'+os.path.splitext(os.path.basename(revfastqfile))[0]
			datasetname   = datasetname.replace('_', '-')+ '-'  + str(grp)

			if options.mergeBAM != "yes":
				bam_file_name =  'primary_%s_%s_visible_bam' % ( options.outputID, datasetname)
				bam_out_name  =  os.path.join( options.outputDIR, bam_file_name)
			else:
				bam_file_name = datasetname
				bam_out_name  =  os.path.join( tmp_dir, bam_file_name)
				readGroup = '@RG\tID:%s\tSM:%s' % ( bam_file_name, IndivName )
				rgfile.write(readGroup+"\n")
			
			result.append(bam_out_name)

			tmp_align_out2 = tempfile.NamedTemporaryFile( dir=tmp_dir )
			tmp_align_out2_name = tmp_align_out2.name
			tmp_align_out2.close()

			command = gen_alignment_cmds 
			
			cmd2.append( 'bwa aln %s %s %s > %s' % ( aligning_cmds, ref_file_name, forfastqfile, tmp_align_out_name ) )
			cmd2b.append( 'bwa aln %s %s %s > %s' % ( aligning_cmds, ref_file_name, revfastqfile, tmp_align_out2_name ) )
			if options.do_remove_unmapped=="yes":
				cmd3.append( 'bwa sampe %s %s %s %s %s %s | samtools view -bSh -F 4 - >> %s' % ( command, ref_file_name, tmp_align_out_name, tmp_align_out2_name, forfastqfile, revfastqfile, bam_out_name ) )
			else:	
				cmd3.append( 'bwa sampe %s %s %s %s %s %s | samtools view -bSh - >> %s' % ( command, ref_file_name, tmp_align_out_name, tmp_align_out2_name, forfastqfile, revfastqfile, bam_out_name ) )			
			grp = grp + 1
	    
	if options.mergeBAM == "yes":
		rgfile.close()					
		
	# perform alignments
	buffsize = 1048576
	try:
		# need to nest try-except in try-finally to handle 2.4
		try:
			# align
			try:
				print("BWA aln execution (forward)\n")
				tmp = tempfile.NamedTemporaryFile( dir=tmp_dir ).name
				
				# Execute BWA aln (forward)
				tmp_stderr = open( tmp, 'wb' )
				for cmdx in cmd2:
					print (' -> BWA aln (forward) will be executed with the following command line: ' + cmdx)
					proc = subprocess.Popen( args=cmdx, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno() )
					returncode = proc.wait()
					if returncode != 0:
						print ("  --> A fatal error occured during the execution procedure..\n")
						raise Exception, stderr
					else:
						print ("  --> Execution complete !\n")
				tmp_stderr.close()
				
				# get stderr, allowing for case where it's very large
				tmp_stderr = open( tmp, 'rb' )
				stderr = ''
				try:
					while True:
						stderr += tmp_stderr.read( buffsize )
						if not stderr or len( stderr ) % buffsize != 0:
							break
				except OverflowError:
					pass
				tmp_stderr.close()

			except Exception, e:
				raise Exception, 'Error mapping sequence. ' + str( e )
			
			# and again if paired data
			if options.genAlignType == 'paired':				
				try:
					print("BWA aln execution (reverse / second member of the pair)\n")
					if cmd2b:
						tmp = tempfile.NamedTemporaryFile( dir=tmp_dir ).name
						
						# Execute BWA aln (reverse)
						tmp_stderr = open( tmp, 'wb' )
						for cmdx in cmd2b:
							print (' -> BWA aln (reverse) will be executed with the following command line: ' + cmdx)
							proc = subprocess.Popen( args=cmdx, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno() )
							returncode = proc.wait()
							if returncode != 0:
								print ("  --> A fatal error occured during the execution procedure..\n")
								raise Exception, stderr
							else:
								print ("  --> Execution complete !\n")
						tmp_stderr.close()
						
						# get stderr, allowing for case where it's very large
						tmp_stderr = open( tmp, 'rb' )
						stderr = ''
						try:
							while True:
								stderr += tmp_stderr.read( buffsize )
								if not stderr or len( stderr ) % buffsize != 0:
									break
						except OverflowError:
							pass
						tmp_stderr.close()

				except Exception, e:
					raise Exception, 'Error mapping reverse Reads. ' + str( e )
			
			# generate align
			try:
				print("BWA samse/sampe execution\n")
				tmp = tempfile.NamedTemporaryFile( dir=tmp_dir ).name
				
				# Execute BWA samse/sampe
				tmp_stderr = open( tmp, 'wb' )
				for cmdx in cmd3:
					print (' -> BWA samse/sampe will be executed with the following command line: ' + cmdx)
					proc = subprocess.Popen( args=cmdx, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno() )
					returncode = proc.wait()
					if returncode != 0:
						print ("  --> A fatal error occured during the execution procedure..\n")
						raise Exception, stderr
					else:
						print ("  --> Execution complete !\n")
				tmp_stderr.close()
				
				# get stderr, allowing for case where it's very large
				tmp_stderr = open( tmp, 'rb' )
				stderr = ''
				try:
					while True:
						stderr += tmp_stderr.read( buffsize )
						if not stderr or len( stderr ) % buffsize != 0:
							break
				except OverflowError:
					pass
				tmp_stderr.close()

			except Exception, e:
				raise Exception, 'Error generating mapping results. ' + str( e ) 
			
			if options.mergeBAM == "yes":
				listebamfiles = ""
				for tfile in result:
					listebamfiles = listebamfiles + " " + tfile 
				#extract a header from one bam
				tmp_out = tempfile.NamedTemporaryFile( dir=tmp_dir)
				tmp_out_name = tmp_out.name
				tmp_out.close()
				cmd = "cat %s > %s; samtools view -H %s >> %s"%(options.outputGRP, tmp_out_name, result[0], tmp_out_name)
				runCmd(cmd, tmp_dir)
				
				cmd = "samtools merge  -rf -h %s %s %s"%(tmp_out_name , options.output,  listebamfiles)
				runCmd(cmd, tmp_dir)
					
			
	
			if options.mergeBAM == "yes":
				# check that there are results in the output file
				if os.path.getsize( options.output ) > 0:
					print ("BWA execution done !\n")
				else:
					raise Exception, 'The output file is empty. You may simply have no matches, or there may be an error with your input file or settings.'
		except Exception, e:
			stop_err( 'The mapping failed.\n' + str( e ) )

	finally:
		# clean up temp dir
		if os.path.exists( tmp_index_dir ):
			shutil.rmtree( tmp_index_dir )
		if os.path.exists( tmp_dir ):
			shutil.rmtree( tmp_dir )
	
	print "\nStop time: ", strftime("%d-%m-%Y_%H:%M:%S", localtime())
	print("################################################################")
	print("#                        End                                   #")
	print("################################################################\n")

if __name__=="__main__": __main__()
