#!/usr/bin/perl

## Date and Version
my $version= '1.5';
my $version_date = '30_jan_2011';

##################################################
## Documentation POD
##################################################

##################################################
## Modules
##################################################
## Perl modules
use strict;
use warnings;
use diagnostics;

use Getopt::Long;
use File::Basename;
use File::Copy;
use Switch;

##################################################
## Globals variables
##################################################

# Initializations
my $Start_time_hr = humanReadableDate();
my $help = undef();
my $DEBUG = 0;
my %Parameters;

# Default values
my $Default_gapN_site = 0.5;
my $Default_gapN_seq = 0.5;
my $Default_min_nb_cod = 10;
my $Default_stop_tol = 20;
my $Default_int_codon = 0;
my $Default_calc = 1;
my $Default_header = 0;
my $Default_verbose = 0;


##################################################
## Manage options
##################################################

# Get options from command line
GetOptions (	'help|h|?'				=> \$help,
			'pfasfile|pfas=s'			=> \$Parameters{'PFAS_file'},
			'treefile|tree=s'			=> \$Parameters{'Tree_file'},
			'focal_species|species=s'	=> \$Parameters{'Focal_species_name'},
			'gapN_site|nsite=f'		=> \$Parameters{'Max_N_by_site'},
			'gapN_seq|nseq=f'			=> \$Parameters{'Max_N_by_sequence'},
			'min_codon|min=i'			=> \$Parameters{'Min_complete_codon'},
			'stop_tolerance|stop=i'	=> \$Parameters{'Stop_position_tolerance'},
			'internal_stop|intstop:i'	=> \$Parameters{'Internal_stop_codon'},
			'make_calculation|calc:i'	=> \$Parameters{'Make_calculation'},
			'print_header|header:i'	=> \$Parameters{'Print_header'},
			'verbose|v:i'				=> \$Parameters{'dNdSpiNpiS_verbose_level'},
			'clean|del=s'				=> \$Parameters{'clean_tmp_files'},
			'summary|sum=s'			=> \$Parameters{'Display_summary'},
			'output|out:s'			=> \$Parameters{'Main_output'},
			'exec:s'					=> \$Parameters{'Execution_mode'}
		);

# Display help message if needed
displayHelpMessage() if $help;

# Check parameters
checkParameters(\%Parameters);

# Generate additional output file name
generateOutputfileNames(\%Parameters);

##################################################
## Main code
##################################################

# Welcome message
print "############################################################\n";
print "# Welcome in Multi dNdSpiNpiS Galaxy Wrapper (Version $version) #\n"; 
print "############################################################\n";
print "Start time: " . $Start_time_hr . "\n\n";

# Initializations
my $Contig_counter = 0;
my ($Current_file_content, $Current_contig_name) = ("", "");
my %Status = ( 'Ok' => 0, 'No_output' => 0, 'Warning' => 0, 'Segmentation_fault' => 0, 'Error' => 0);

# Verbose
print 'Selected Fasta/PseudoFasta input file: ' . $Parameters{'PFAS_file'} . "\n";
print 'Selected Tree file (newick format): ' . $Parameters{'Tree_file'} . "\n\n";

print 'Selected focal species: ' . $Parameters{'Focal_species_name'} . "\n\n";

print "Multi fasta file analysis procedure:\n\n";

# Launch a dNdSpiNpiS instance for each contig of the clean Fasta/PseudoFasta file
open (MAIN_PFAS, '<' . $Parameters{'PFAS_file'}) or die ('Error: Cannot open/read file: ' . $Parameters{'PFAS_file'} . "\n");

while (my $Current_line = <MAIN_PFAS>){
	
	if ($Current_line =~ /^[^#|@]/) { # If the current line ins't a comment line we analyse it
		
		if ($Current_line =~ /^>([\w\.]+)\|/) {
			
			if ($Contig_counter > 0 && $1 ne $Current_contig_name) {
				
				processContig($Current_file_content, $Contig_counter, $Current_contig_name, \%Parameters, \%Status);
				
				# Reinitializations
				$Current_file_content = "";
			}
			
			# Increase the number of treated contigs
			if ($1 ne $Current_contig_name) { $Contig_counter++; }
			
			# Store the new contig name for the next processContig call
			$Current_contig_name = $1;
		}
		
		$Current_file_content .= $Current_line;
	}
}

# Process the last contig of the input file
if ($Contig_counter > 0) {
	processContig($Current_file_content, $Contig_counter, $Current_contig_name, \%Parameters, \%Status);
} else {
	print 'Error: No contigs found in the Fasta/PseudoFasta input file !' . "\n\n";
	exit(1);
}

close (MAIN_PFAS);

print 'Global output file: ' . $Parameters{'Main_output'} . "\n\n\n";

# Display a summary of the execution (Informatic point of view)
print 'Execution statistics:' . "\n\n";
foreach my $Current_status (keys (%Status)) {
	print ' - ' . $Current_status . ': ' . $Status{$Current_status} . "\n";
}
print "\n\n";

# Display a summary of the execution (Informatic point of view)
if ($Parameters{'Display_summary'} eq 'yes') {
	if (-e $Parameters{'Main_output'} && ! -z $Parameters{'Main_output'}) {
		if ($Parameters{'dNdSpiNpiS_verbose_level'} == 0) {
			print 'Biological summary of the analysis:' . "\n";
			execute_dNdSpiNpiS_summary(\%Parameters);
		} else {
			print 'Sorry, dNdSpiNpiS_summary is incompatible with dNdSpiNpiS internal verbose mode !' . "\n";
		}
	} else {
		print 'Warning: dNdSpiNpiS Global output file does not exist ! dNdSpiNpiS_summary won\'t be executed !' . "\n";
	}
}

print "\nStop time: " . humanReadableDate() . "\n";
print "############################################################\n";
print "#                      End of execution                    #\n"; 
print "############################################################\n";

##################################################
## Filename management related functions
##################################################

sub generateOutputfileNames {
	
	# Recovers parameters
	my $Ref_to_parameters = shift;
	
	# Some explanations: 
	# In local mode the names of the additionnal output files are derived from the name of the main input file
	# However, the current version of Galaxy Workflow Manager is not able to retrieve output files with non HARD CODED names (The value of the from_work_dir attribute is not interpreted !!)
	# Therefore it's impossible to use derived name in Galaxy mode
	
	if ($Ref_to_parameters->{'Execution_mode'} eq 'local') {
		# Names derived from the main input file name
		my $Input_file_basename = basename($Ref_to_parameters->{'PFAS_file'}) ;
		$Ref_to_parameters->{'Summary_file'} = $Input_file_basename . '.summary';
		$Ref_to_parameters->{'Main_error'} = $Input_file_basename . '.global_error';		
		$Ref_to_parameters->{'Contig_no_output'} = $Input_file_basename . '.no_output_contigs';
		$Ref_to_parameters->{'Contig_with_warning'} = $Input_file_basename .  '.warning_contigs';
		$Ref_to_parameters->{'Contig_with_error'} = $Input_file_basename . '.error_contigs';
		$Ref_to_parameters->{'Contig_with_segfault'} = $Input_file_basename . '.segfault_contigs';
	} else {
		# Hard coded names (Galaxy mode)
		$Ref_to_parameters->{'Summary_file'} = 'dNdSpiNpiS_summary';
		$Ref_to_parameters->{'Main_error'} = 'dNdSpiNpiS_error';
		$Ref_to_parameters->{'Contig_no_output'} = 'dNdSpiNpiS_no_output_contigs';
		$Ref_to_parameters->{'Contig_with_warning'} = 'dNdSpiNpiS_warning_contigs';
		$Ref_to_parameters->{'Contig_with_error'} = 'dNdSpiNpiS_error_contigs';
		$Ref_to_parameters->{'Contig_with_segfault'} = 'dNdSpiNpiS_segfault_contigs';
	}		
	
	return 0;
}

##################################################
## Contig processing related functions
##################################################

sub processContig {
	
	# Recovers parameters
	my ($File_content, $Contig_number, $Contig_name, $Ref_to_parameters, $Ref_to_status) = @_;
	
	# Initializations
	$Ref_to_parameters->{'dNdSpiNpiS_input_file'} = basename($Ref_to_parameters->{'PFAS_file'}) . '.contig_' . $Contig_number;
	$Ref_to_parameters->{'dNdSpiNpiS_stdout_file'} = $Ref_to_parameters->{'dNdSpiNpiS_input_file'} . '.pop';
	$Ref_to_parameters->{'dNdSpiNpiS_error_file'} = $Ref_to_parameters->{'dNdSpiNpiS_input_file'} . '.err';
	$Ref_to_parameters->{'dNdSpiNpiS_output_file'} = $Ref_to_parameters->{'dNdSpiNpiS_stdout_file'};
	
	print ' Analysis of population genetics for contig number ' . $Contig_number . ' (Original name: ' . $Contig_name . ")\n";
				
	# Real processing
	create_dNdSpiNpiS_InputFile($File_content, $Ref_to_parameters);
	execute_dNdSpiNpiS($Ref_to_parameters, $Contig_number);
	
	# Determine output status
	my $Output_status = checkOutputFiles($Ref_to_parameters->{'dNdSpiNpiS_output_file'}, $Ref_to_parameters->{'dNdSpiNpiS_error_file'});

	# Display output status
	dealWithOutputStatus($Output_status, $Contig_number, $Contig_name, $Ref_to_status, $Ref_to_parameters);
	
	# Remove temporary files if requested
	if ($Ref_to_parameters->{'clean_tmp_files'} eq 'yes') {
		if ($Output_status ne "Warning" && $Output_status ne "Error" && $Output_status ne "Segmentation_fault") {
			cleanItAfterMe($Ref_to_parameters->{'dNdSpiNpiS_input_file'}, $Ref_to_parameters->{'dNdSpiNpiS_output_file'}, $Ref_to_parameters->{'dNdSpiNpiS_stdout_file'}, $Ref_to_parameters->{'dNdSpiNpiS_error_file'});
		}
	}
	
	return 0;
}

sub create_dNdSpiNpiS_InputFile {
	
	# Recovers parameters
	my ($GENO_informations, $Ref_to_parameters) = @_;
	
	# Remove old file if exists
	if (-e $Ref_to_parameters->{'dNdSpiNpiS_input_file'}) { unlink($Ref_to_parameters->{'dNdSpiNpiS_input_file'}); }
	
	# New file creation
	open (DNDS_INPUT, '>' . $Ref_to_parameters->{'dNdSpiNpiS_input_file'}) or die ('Error: Cannot create/open file: ' . $Ref_to_parameters->{'dNdSpiNpiS_input_file'} . "\n");
	print DNDS_INPUT $GENO_informations;
	close (DNDS_INPUT);
	
	return 0;
}

##################################################
## Tools execution related functions
##################################################

sub execute_dNdSpiNpiS {
	
	# Recovers parameters
	my ($Ref_to_parameters, $Current_contig_number) = @_;
	
	# Command line building
	my $cmd = 'dNdSpiNpiS ' . $Ref_to_parameters->{'dNdSpiNpiS_input_file'} . ' ' . $Ref_to_parameters->{'Tree_file'};
	$cmd .= ' ' . $Ref_to_parameters->{'Focal_species_name'};
	$cmd .= ' ' . $Ref_to_parameters->{'Max_N_by_site'};
	$cmd .= ' ' . $Ref_to_parameters->{'Max_N_by_sequence'};
	$cmd .= ' ' . $Ref_to_parameters->{'Min_complete_codon'};
	$cmd .= ' ' . $Ref_to_parameters->{'Stop_position_tolerance'};
	$cmd .= ' ' . $Ref_to_parameters->{'Internal_stop_codon'};
	$cmd .= ' ' . $Ref_to_parameters->{'Make_calculation'};
	if ($Current_contig_number == 1) { $cmd .= ' 1'; } else { $cmd .= ' 0'; } # Header
	$cmd .= ' ' . $Ref_to_parameters->{'dNdSpiNpiS_verbose_level'};
	$cmd .= ' 1> ' . $Ref_to_parameters->{'dNdSpiNpiS_stdout_file'} . ' 2> ' . $Ref_to_parameters->{'dNdSpiNpiS_error_file'};

	print '  -> dNdSpiNpiS will be executed with the following command line: ' . $cmd . "\n" if $DEBUG;

	system ($cmd);
	
	return 0;
}

sub execute_dNdSpiNpiS_summary {

	# Recovers parameters
	my ($Ref_to_parameters) = shift;
	
	# Rename an already existing summary file if needed
	if (-e $Ref_to_parameters->{'Summary_file'}) { move($Ref_to_parameters->{'Summary_file'}, $Ref_to_parameters->{'Summary_file'} . '.old.' . $Start_time_hr); }
	
	# Command line building
	my $cmd = 'dNdSpiNpiS_summary ' . $Ref_to_parameters->{'Main_output'} . ' ' . $Ref_to_parameters->{'Focal_species_name'} . ' 1> ' . $Ref_to_parameters->{'Summary_file'} . ' 2>&1';
	print ' -> dNdSpiNpiS_summary will be executed with the following command line: ' . $cmd . "\n" if $DEBUG;

	# Execution
	system ($cmd);

	# Display dNdSpiNpiS summary log 
	open (SUMMARY, '<' . $Ref_to_parameters->{'Summary_file'}) or die ('Error: Cannot open/read file: ' . $Ref_to_parameters->{'Summary_file'} . "\n");
	while (my $summary_line = <SUMMARY>){
		print $summary_line;
	}				
	close (SUMMARY);	
	
	return 0;
}

##################################################
## Output mangement related functions
##################################################

sub dealWithOutputStatus {
	
	# Recovers parameters
	my ($New_status, $Contig_number, $Contig_name, $Ref_to_status, $Ref_to_parameters) = @_;
	
	switch ($New_status) {
		case "Ok" {
			print '   --> Execution complete !!' . "\n\n"; 
			# Add unitary result to the global output file
			addDataToGlobalOutputFile($Ref_to_parameters->{'dNdSpiNpiS_output_file'}, $Ref_to_parameters->{'Main_output'}, $Contig_number);
		}
		case "No_output" {
			print '   --> There are no dNdSpiNpiS results for contig ' . $Contig_number . ' (' . $Contig_name . ")..\n\n";
			# Regroup contigs for which there is no normal output (for relaunch)
			addDataToGlobalOutputFile($Ref_to_parameters->{'dNdSpiNpiS_input_file'}, $Ref_to_parameters->{'Contig_no_output'}, $Contig_number) if ($Ref_to_parameters->{'Execution_mode'} eq 'local');
		}
		case "Warning" {
			print '   --> Execution complete (with warnings) ! Results might be uncomplete for contig ' . $Contig_number . ' (' . $Contig_name . ")..\n\n";
			# Add unitary result to the global output file
			addDataToGlobalOutputFile($Ref_to_parameters->{'dNdSpiNpiS_output_file'}, $Ref_to_parameters->{'Main_output'}, $Contig_number); 
			# Regroup contigs that generate warnings (for relaunch)
			addDataToGlobalOutputFile($Ref_to_parameters->{'dNdSpiNpiS_input_file'}, $Ref_to_parameters->{'Contig_with_warning'}, $Contig_number) if ($Ref_to_parameters->{'Execution_mode'} eq 'local');
			# Add new warnings to the global error file
			addDataToGlobalOutputFile($Ref_to_parameters->{'dNdSpiNpiS_error_file'}, $Ref_to_parameters->{'Main_error'}, $Contig_number, '## Warning(s) for contig ' . $Contig_number . ' (' . $Contig_name . ') ##'); 
		}
		case "Segmentation_fault" {
			print '   --> A segmentation fault was raised during the processing of contig ' . $Contig_number . ' (' . $Contig_name . ")..\n\n";
			# Regroup contigs that generate segmentation faults (for relaunch)
			addDataToGlobalOutputFile($Ref_to_parameters->{'dNdSpiNpiS_input_file'}, $Ref_to_parameters->{'Contig_with_segfault'}, $Contig_number) if ($Ref_to_parameters->{'Execution_mode'} eq 'local');
			# Add new segfaults to the global error file
			addDataToGlobalOutputFile($Ref_to_parameters->{'dNdSpiNpiS_error_file'}, $Ref_to_parameters->{'Main_error'}, $Contig_number, '## Segmentation fault for contig ' . $Contig_number . ' (' . $Contig_name . ') ##'); 
		}
		case "Error" {
			print '   --> A fatal error occured during the processing of contig ' . $Contig_number . ' (' . $Contig_name . ")..\n\n";
			# Regroup contigs that generate error messages (for relaunch)
			addDataToGlobalOutputFile($Ref_to_parameters->{'dNdSpiNpiS_input_file'}, $Ref_to_parameters->{'Contig_with_error'}, $Contig_number) if ($Ref_to_parameters->{'Execution_mode'} eq 'local');
			# Add new errors to the global error file
			addDataToGlobalOutputFile($Ref_to_parameters->{'dNdSpiNpiS_error_file'}, $Ref_to_parameters->{'Main_error'}, $Contig_number, '## Error(s) for contig ' . $Contig_number . ' (' . $Contig_name . ') ##'); 
		}
		else { print "Unexpected switch case !  !" }
	}
	
	$Ref_to_status->{$New_status}++;
	
	return 0;
}

sub addDataToGlobalOutputFile {
	
	# Recovers parameters
	my ($New_data, $Global_file, $Counter, $Opt_header) = @_;
	
	# Rename an already existing global file / Special Contig file if needed
	if (-e $Global_file && $Counter == 1) { move($Global_file, $Global_file . '.old.' . $Start_time_hr); }
	
	# Add the content of the new result file to the global result file
	open(GLOBAL, '>>' . $Global_file) or die ('Error: Cannot create/open file: ' . $Global_file . "\n");
	
	if ($Opt_header) { print GLOBAL $Opt_header . "\n"; }
	
	open (DATA, '<' . $New_data) or die ('Error: Cannot open/read file: ' . $New_data . "\n");	
	while (my $line = <DATA>){
		print GLOBAL $line;
	}
	close (DATA);
	
	close(GLOBAL);

	return 0;
}

##################################################
## Help/Check related functions
##################################################

sub displayHelpMessage {

	print('##################################################' . "\n");
	print('#    dNdSpiNpiS Galaxy Wrapper - Help section    #' . "\n"); 
	print('##################################################' . "\n\n");

	print "Usage example (short names): dNdSpiNpiS_wrapper.pl -pfas sample.pfas_clean -tree sample_newick.tree -species My_species_name -nsite " . $Default_gapN_site . " -nseq " . $Default_gapN_seq . " -min " . $Default_min_nb_cod . " -stop " . $Default_stop_tol . " -intstop " . $Default_int_codon . " -calc " . $Default_calc . " -header " . $Default_header . " -v " . $Default_verbose . " -del yes -sum yes -out sample.pop\n\n";
	
	print "Usage example (long names): dNdSpiNpiS_wrapper.pl -pfasfile sample.pfas_clean -treefile sample_newick.tree -focal_species My_species_name -gapN_site " . $Default_gapN_site . " -gapN_seq " . $Default_gapN_seq . " -min_codon " . $Default_min_nb_cod . " -stop_tolerance " . $Default_stop_tol . " -internal_stop " . $Default_int_codon . " -make_calculation " . $Default_calc . " -print_header " . $Default_header . " -verbose " . $Default_verbose . " -clean yes -summary yes -out sample.pop\n\n";
	
	print "Here is the list of authorized parameters :\n\n";
	
	print " -help => Display this help message.\n\n";
	
	print " -pfasfile/-pfas file => Path and name of the input file in Fasta/PseudoFasta format (mandatory).\n\n";
	
	print " -treefile/-tree file => Path and name of the input tree file in Newick format (mandatory).\n\n";
	
	print " -focal_species/-species string => Name of the focal species for the current study (mandatory).\n\n";
	
	print " -gapN_site/-nsite positive float => Max tolerated proportion of undetermined characters for a site (ie. a codon of the alignment) (optional). Range of possible values [0 to 1].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_gapN_site . "\n\n";
	
	print " -gapN_seq/-nseq positive float => Max tolerated proportion of undetermined characters for a sequence (optional). Range of possible values [0 to 1].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_gapN_seq . "\n\n";
	
	print " -min_codon/-min integer => Minimum number of complete codons required to perform calculations (optional). Possible values are [All positive integers].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_min_nb_cod . "\n\n";
	
	print " -stop_tolerance/-stop integer => Number of terminal positions in which stop codons are tolerated (optional). Possible values are [All positive integers].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_stop_tol . "\n\n";
	
	print " -internal_stop/-intstop boolean => Define if internal stop codon (ie. codons between the two tolerance zone) are tolerated and will be replaced by NNN (1), or forbidden (0). Possible values are [0|1].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_int_codon . "\n\n";
	
	print " -make_calculation/-calc boolean => Define if calculations must be done and output (1), or not (0) (optional). Possible values are [0|1].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_calc . "\n\n";
	
	print " -print_header/-header boolean => Define if the header must be output (1), or not (0) (optional). Possible values are [0|1].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_header . "\n\n";
	
	print " -verbose/-v boolean => Define if dNdSpiNpiS must be wordy (1), or not (0) (Incompatible with the summary option (-summary/-sum)) (optional). Possible values are [0|1].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_verbose . "\n\n";	
	
	print " -clean/-del string => Activate/Desactivate temporary files removal (optional). Possible values are [yes|no].\n";
	print "\tBy default, temporary files will not be conserved for execution instance without error.\n\n";
	
	print " -summary/-sum string => Activate/Desactivate biological summary display (Incompatible with dNdSpiNpiS internal verbose option (-verbose/-v)) (optional). Possible values are [yes|no].\n";
	print "\tBy default, the biological summary of the analysis will be added at the end of the log file.\n\n";	
	
	print " -output/-out string => Path and name of the final output file (optional).\n";
	print "\tIf not specified, a default name (derived from the name of the Fasta/PseudoFasta input file) will be used.\n\n";

	exit (1);
}

sub checkParameters {
	
	# Recovers parameters
	my $Ref_to_parameters = shift;
	
	# Execution mode
	if (!defined($Ref_to_parameters->{'Execution_mode'}) || $Ref_to_parameters->{'Execution_mode'} eq "") {
		$Ref_to_parameters->{'Execution_mode'} = 'local';
	}
	
	# Deal with global options (Input, Output and cleaning)
	if (!defined($Ref_to_parameters->{'PFAS_file'})) {
		print 'Error: No Fasta/PseudoFasta input file defined !' . "\n\n";
		displayHelpMessage();
	} else {
		checkInputFile($Ref_to_parameters->{'PFAS_file'}, 'Fasta/PseudoFasta');
	}
	
	if (!defined($Ref_to_parameters->{'Tree_file'})) {
		print 'Error: No Tree file defined !' . "\n\n";
		displayHelpMessage();
	} else {
		checkInputFile($Ref_to_parameters->{'Tree_file'}, 'Tree');
	}
	
	if (!defined($Ref_to_parameters->{'Main_output'}))  {
		print 'Warning: The name of the main output file has been omited ! A default name (derived from the name of the Fasta/PseudoFasta input file) will be used !' . "\n\n";	
		$Ref_to_parameters->{'Main_output'} = basename($Ref_to_parameters->{'PFAS_file'}) .  '.pop';
	}

	if (!defined($Ref_to_parameters->{'clean_tmp_files'}))  {
		print 'Warning: The clean/del parameter has been omited ! Temporary files will not be conserved !' . "\n\n";	
		$Ref_to_parameters->{'clean_tmp_files'} = 'yes';
		
	} elsif ($Ref_to_parameters->{'clean_tmp_files'} !~ /^(yes|no)$/i) {
		print 'Warning: The value (' . $Ref_to_parameters->{'clean_tmp_files'} . ') of the clean/del parameter given to this module is not valid ! Temporary files will not be conserved !' . "\n\n";
		$Ref_to_parameters->{'clean_tmp_files'} = 'yes';
	}

	if (!defined($Ref_to_parameters->{'Display_summary'}))  {
		print 'Warning: The summary/sum parameter has been omited ! Biological summary will be added at the end of the log file !' . "\n\n";	
		$Ref_to_parameters->{'Display_summary'} = 'yes';
		
	} elsif ($Ref_to_parameters->{'Display_summary'} !~ /^(yes|no)$/i) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Display_summary'} . ') of the summary/sum parameter given to this module is not valid ! Biological summary will be added at the end of the log file !' . "\n\n";
		$Ref_to_parameters->{'Display_summary'} = 'yes';
	}
	
	# Deal with dNdSpiNpiS specific options
	if (!defined($Ref_to_parameters->{'Focal_species_name'})) {
		print 'Error: No focal species name specified !' . "\n\n";
		displayHelpMessage();
	}	
	
	if (!defined($Ref_to_parameters->{'Max_N_by_site'})) {
		print 'Warning: No "gapN_site|nsite" parameter specified ! Default value ("' . $Default_gapN_site . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Max_N_by_site'} = $Default_gapN_site;
		
	} elsif ($Ref_to_parameters->{'Max_N_by_site'} < 0 || $Ref_to_parameters->{'Max_N_by_site'} > 1) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Max_N_by_site'} . ') of the "gapN_site|nsite" parameter given to this module is not a valid numeric value between 0 and 1! Default value ("' . $Default_gapN_site . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Max_N_by_site'} = $Default_gapN_site;
	}
	
	if (!defined($Ref_to_parameters->{'Max_N_by_sequence'})) {
		print 'Warning: No "gapN_seq|nseq" parameter specified ! Default value ("' . $Default_gapN_seq . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Max_N_by_sequence'} = $Default_gapN_seq;
		
	} elsif ($Ref_to_parameters->{'Max_N_by_sequence'} < 0 || $Ref_to_parameters->{'Max_N_by_sequence'} > 1) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Max_N_by_sequence'} . ') of the "gapN_seq|nseq" parameter given to this module is not a valid numeric value between 0 and 1! Default value ("' . $Default_gapN_seq . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Max_N_by_sequence'} = $Default_gapN_seq;
	}
	
	if (!defined($Ref_to_parameters->{'Min_complete_codon'})) {
		print 'Warning: No "min_codon|min" parameter specified ! Default value ("' . $Default_min_nb_cod . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Min_complete_codon'} = $Default_min_nb_cod;
		
	} elsif ($Ref_to_parameters->{'Min_complete_codon'} !~ /^[0-9]+$/ || $Ref_to_parameters->{'Min_complete_codon'} <= 0) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Min_complete_codon'} . ') of the "min_codon|min" parameter given to this module is not valid ! Default value ("' . $Default_min_nb_cod . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Min_complete_codon'} = $Default_min_nb_cod;
	}
	
	if (!defined($Ref_to_parameters->{'Stop_position_tolerance'})) {
		print 'Warning: No "stop_tolerance|stop" parameter specified ! Default value ("' . $Default_stop_tol . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Stop_position_tolerance'} = $Default_stop_tol;
		
	} elsif ($Ref_to_parameters->{'Stop_position_tolerance'} !~ /^[0-9]+$/ || $Ref_to_parameters->{'Stop_position_tolerance'} <= 0) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Stop_position_tolerance'} . ') of the "stop_tolerance|stop" parameter given to this module is not valid ! Default value ("' . $Default_stop_tol . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Stop_position_tolerance'} = $Default_stop_tol;
	}
	
	if (!defined($Ref_to_parameters->{'Internal_stop_codon'})) {
		print 'Warning: No "internal_stop|intstop" parameter specified ! Default value ("' . $Default_int_codon . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Internal_stop_codon'} = $Default_int_codon;
		
	} elsif ($Ref_to_parameters->{'Internal_stop_codon'} !~ /^[0-1]$/) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Internal_stop_codon'} . ') of the "internal_stop|intstop" parameter given to this module is not valid ! Default value ("' . $Default_int_codon . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Internal_stop_codon'} = $Default_int_codon;
	}

	if (!defined($Ref_to_parameters->{'Make_calculation'})) {
		print 'Warning: No "make_calculation|calc" parameter specified ! Default value ("' . $Default_calc . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Make_calculation'} = $Default_calc;
		
	} elsif ($Ref_to_parameters->{'Make_calculation'} !~ /^[0-1]$/) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Make_calculation'} . ') of the "make_calculation|calc" parameter given to this module is not valid ! Default value ("' . $Default_calc . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Make_calculation'} = $Default_calc;
	}

	if (!defined($Ref_to_parameters->{'Print_header'})) {
		print 'Warning: No "print_header|header" parameter specified ! Default value ("' . $Default_header . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Print_header'} = $Default_header;
		
	} elsif ($Ref_to_parameters->{'Print_header'} !~ /^[0-1]$/) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Print_header'} . ') of the "print_header|header" parameter given to this module is not valid ! Default value ("' . $Default_header . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Print_header'} = $Default_header;
	}
	
	if (!defined($Ref_to_parameters->{'dNdSpiNpiS_verbose_level'})) {
		print 'Warning: No "verbose|v" parameter specified ! Default value ("' . $Default_verbose . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'dNdSpiNpiS_verbose_level'} = $Default_verbose;
		
	} elsif ($Ref_to_parameters->{'dNdSpiNpiS_verbose_level'} !~ /^[0-1]$/) {
		print 'Warning: The value (' . $Ref_to_parameters->{'dNdSpiNpiS_verbose_level'} . ') of the "verbose|v" parameter given to this module is not valid ! Default value ("' . $Default_verbose . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'dNdSpiNpiS_verbose_level'} = $Default_verbose;
	}
		
	return 0;
}

sub checkInputFile {
	
	# Recovers parameters
	my ($file, $type) = @_;
	
	if (! -e $file || -z $file) {
		print 'Fatal error: The ' . $type . ' file specified in the command line does not exist or is empty ! Please check file name, path and content !!' . "\n";
		print 'Selected file: ' . basename($file) . "\n";
		print 'Corresponding directory: ' . dirname($file) . "\n\n";
		exit(1);
	}
	
	return 0;
}

sub checkOutputFiles {
	
	# Recovers parameters
	my ($Output_file, $Error_file) = @_;
	
	# Initializations
	my $Status = 'Unknown';

	if (-e $Error_file && ! -z $Error_file) {
		
		# Check for critical errors
		open (ERROR_FILE, '<' . $Error_file) or die ('Error: Cannot open/read file: ' . $Error_file . "\n");
		while (my $error_line = <ERROR_FILE>){
			if ($error_line =~ /Segmentation[\s_]*fault|segfault|Erreur de segmentation/i) { $Status = 'Segmentation_fault'; }
		}				
		close (ERROR_FILE);
		
		if ($Status ne 'Segmentation_fault') {
			
			if ( -e $Output_file && ! -z $Output_file) {
				$Status = 'Warning';
			} else {
				$Status = 'Error';
			}
		}
		
	} else {
		
		if ( -e $Output_file && ! -z $Output_file) {
			$Status = 'Ok';
		} else {
			$Status = 'No_output';
		}
	}
	
	return $Status;
}

##################################################
## Basic common functions
##################################################

sub humanReadableDate {
	
	# Recovers parameters
	my $time = shift || time;
	
	# Split time string
	my ($seconde, $minute, $heure, $jour, $mois, $annee, $jour_semaine, $jour_annee, $heure_hiver_ou_ete) = localtime($time);
	$mois  += 1;
	$annee += 1900;

	# On rajoute 0 si le chiffre est compris entre 1 et 9
	foreach ( $seconde, $minute, $heure, $jour, $mois, $annee ) { s/^(\d)$/0$1/; }
	
	return "$jour-$mois-$annee" . '_' . "$heure:$minute:$seconde";
}

sub cleanItAfterMe {
	
	# Remove all files given as arguments to this function
	foreach my $File_to_delete (@_) {
		if (-e  $File_to_delete) {
			unlink($File_to_delete);
		}
	}
	
	return 0;
}
