#!/usr/bin/perl

## Date and Version
my $version= '1.5';
my $version_date = '10-12-2012';

##################################################
## Documentation POD
##################################################

##################################################
## Modules
##################################################
## Perl modules
use strict;
use warnings;
use diagnostics;

use Getopt::Long;
use File::Basename;
use File::Copy;
use Switch;

##################################################
## Globals variables
##################################################

# Initializations
my $Start_time_hr = humanReadableDate();
my $help = undef();
my $DEBUG = 0;
my %Parameters;

# Default values
my $Default_threshold = 0.95;

##################################################
## Manage options
##################################################

# Get options from command line
GetOptions ('help|h'				=> \$help,
			'genofile|geno=s'		=> \$Parameters{'ALR_GEN_file'},
			'threshold|sill=f'		=> \$Parameters{'Geno2pfas_threshold'},
			'clean|del=s'			=> \$Parameters{'clean_tmp_files'},
			'output|out=s'			=> \$Parameters{'PFAS_file'},
			'exec:s'				=> \$Parameters{'Execution_mode'}
		);

# Display help message if needed
if (defined($help)) {
	displayHelpMessage();
}

# Check parameters
checkParameters(\%Parameters);

# Generate additional output file name
generateOutputfileNames(\%Parameters);

# Backup old data
saveOldResults(\%Parameters);

# Galaxy mode only: create all files that will be recovered by Galaxy from_work_dir command
createFromWorkDirFiles(\%Parameters) if ($Parameters{'Execution_mode'} =~ /galaxy/i);

print Dumper(\%Parameters) if $DEBUG;


##################################################
## Main code
##################################################

# Welcome message
print "###########################################################\n";
print "# Welcome in Multi Geno2pfas Galaxy Wrapper (Version $version) #\n"; 
print "###########################################################\n";
print "Start time: " . $Start_time_hr . "\n\n";

# Initializations
my $Contig_counter = 0;
my ($Current_file_content, $Current_contig_name) = ("", "");
my %Status = ( 'Ok' => 0, 'No_output' => 0, 'Warning' => 0, 'Segmentation_fault' => 0, 'Error' => 0);

# Verbose
print 'Selected ALR_GEN input file: ' . $Parameters{'ALR_GEN_file'} . "\n\n";

print "ALR_GEN file analysis procedure:\n\n";

# Launch a Geno2pfas instance for each contig of the global .alr_gen file generated by Reads2SNP
open (MAIN_ALR, '<' . $Parameters{'ALR_GEN_file'}) or die ('Error: Cannot open/read file: ' . $Parameters{'ALR_GEN_file'} . "\n");

while (my $Current_line = <MAIN_ALR>){

	if ($Current_line =~ /^[^#|@]/) { # If the current line ins't a comment line we analyse it
		
		if ($Current_line =~ /^>(\w+)/) {
			
			if ($Contig_counter > 0) {
				
				processContig($Current_file_content, $Contig_counter, $Current_contig_name, \%Parameters, \%Status);
				
				# Reinitializations
				$Current_file_content = "";
			}
			
			# Store the new contig name for the next processContig call
			$Current_contig_name = $1;
			
			# Increase the number of treated contigs
			$Contig_counter++;
		}
		
		$Current_file_content .= $Current_line;
	}
}

# Process the last contig of the input file
processContig($Current_file_content, $Contig_counter, $Current_contig_name, \%Parameters, \%Status);

close (MAIN_ALR);

print 'Global PFAS output file: ' . $Parameters{'PFAS_file'} . "\n\n\n";

# Display a summary of the execution (Informatic point of view)
print 'Execution statistics:' . "\n\n";
foreach my $Current_status (keys (%Status)) {
	print ' - ' . $Current_status . ': ' . $Status{$Current_status} . "\n";
}

print "\nStop time: " . humanReadableDate() . "\n";
print "###########################################################\n";
print "#                     End of execution                    #\n"; 
print "###########################################################\n";

##################################################
## Filename management related functions
##################################################

sub generateOutputfileNames {
	
	# Recovers parameters
	my $Ref_to_parameters = shift;
	
	# Some explanations: 
	# In local mode the names of the additionnal output files are derived from the name of the main input file
	# However, the current version of Galaxy Workflow Manager is not able to retrieve output files with non HARD CODED names (The value of the from_work_dir attribute is not interpreted !!)
	# Therefore it's impossible to use derived name in Galaxy mode
	
	if ($Ref_to_parameters->{'Execution_mode'} eq 'local') {
		# Names derived from the main input file name
		my $Input_file_basename = basename($Ref_to_parameters->{'ALR_GEN_file'}) ;
		$Ref_to_parameters->{'Main_error'} = $Input_file_basename . '.global_error';
		$Ref_to_parameters->{'Contig_no_output'} = $Input_file_basename . '.no_output_contigs';
		$Ref_to_parameters->{'Contig_with_warning'} = $Input_file_basename .  '.warning_contigs';
		$Ref_to_parameters->{'Contig_with_error'} = $Input_file_basename . '.error_contigs';
		$Ref_to_parameters->{'Contig_with_segfault'} = $Input_file_basename . '.segfault_contigs';
	} else {
		# Hard coded names (Galaxy mode)
		$Ref_to_parameters->{'Main_error'} = 'Geno2pfas_error';
		$Ref_to_parameters->{'Contig_no_output'} = 'Geno2pfas_no_output_contigs';
		$Ref_to_parameters->{'Contig_with_warning'} = 'Geno2pfas_warning_contigs';
		$Ref_to_parameters->{'Contig_with_error'} = 'Geno2pfas_error_contigs';
		$Ref_to_parameters->{'Contig_with_segfault'} = 'Geno2pfas_segfault_contigs';
	}		
	
	return 0;
}


##################################################
## Galaxy mode specific functions
##################################################

sub createFromWorkDirFiles {
	
	# Recovers parameters
	my $Ref_to_parameters = shift;
	
	# Some explanations: 
	# In the last versions of Galaxy, an error message is raised when you try to recover a file that does not exists with the from_work_dir command.
	# Therefore, in galaxy mode, all the files requested by a from_work_dir command in the XML file must be created in the Perl wrapper.
	
	if (!-e $Ref_to_parameters->{'Main_error'}) {
		open(GLOBAL_ERROR, '>' . $Ref_to_parameters->{'Main_error'}) or die ('Cannot create/open file: ' . $Ref_to_parameters->{'Main_error'} . "\n");
		close(GLOBAL_ERROR);
	}
}


##################################################
## Backup related functions
##################################################

sub saveOldResults {
	
	# Recovers parameters
	my $Ref_to_parameters = shift;

	# Initializations
	my @list_of_global_file = ('PFAS_file', 'Main_error', 'Contig_no_output', 'Contig_with_warning', 'Contig_with_error', 'Contig_with_segfault');
	
	# Rename an already existing global file / Special Contig file if needed
	foreach my $Global_file_name (@list_of_global_file) {
		rename($Ref_to_parameters->{$Global_file_name}, $Ref_to_parameters->{$Global_file_name} . '.old.' . $Start_time_hr) if (-e $Ref_to_parameters->{$Global_file_name});
	}
	
	return 0;
}


##################################################
## Contig processing related functions
##################################################

sub processContig {
	
	# Recovers parameters
	my ($File_content, $Contig_number, $Contig_name, $Ref_to_parameters, $Ref_to_status) = @_;
	
	# Initializations
	$Ref_to_parameters->{'Geno2pfas_input_file'} = basename($Ref_to_parameters->{'ALR_GEN_file'}) . '.contig_' . $Contig_number;
	$Ref_to_parameters->{'Geno2pfas_stdout_file'} = $Ref_to_parameters->{'Geno2pfas_input_file'} . '.out';
	$Ref_to_parameters->{'Geno2pfas_error_file'} = $Ref_to_parameters->{'Geno2pfas_input_file'} . '.err';
	$Ref_to_parameters->{'Geno2pfas_output_file'} = $Ref_to_parameters->{'Geno2pfas_input_file'} . '.pfas';
	
	print ' Building sequences for contig number ' . $Contig_number . ' (' . $Contig_name . ")\n";
				
	# Real processing
	createGeno2pfasInputFile($File_content, $Ref_to_parameters);
	executeGeno2pfas($Ref_to_parameters);
	
	# Determine output status
	my $Output_status = checkOutputFiles($Ref_to_parameters->{'Geno2pfas_output_file'}, $Parameters{'Geno2pfas_error_file'});
	
	# Display output status
	dealWithOutputStatus($Output_status, $Contig_number, $Contig_name, $Ref_to_status, $Ref_to_parameters);
	
	# Remove temporary files if requested
	if ($Parameters{'clean_tmp_files'} eq 'yes') {
		if ($Output_status ne "Warning" && $Output_status ne "Error" && $Output_status ne "Segmentation_fault") {
			cleanItAfterMe($Parameters{'Geno2pfas_input_file'}, $Ref_to_parameters->{'Geno2pfas_output_file'}, $Parameters{'Geno2pfas_stdout_file'}, $Parameters{'Geno2pfas_error_file'});
		}
	}
	
	return 0;
}

sub createGeno2pfasInputFile {
	
	# Recovers parameters
	my ($GENO_informations, $Ref_to_parameters) = @_;
	
	# Remove old file if exists
	if (-e $Ref_to_parameters->{'Geno2pfas_input_file'}) { unlink($Ref_to_parameters->{'Geno2pfas_input_file'}); }
	
	# New file creation
	open (G2PFAS_INPUT, '>' . $Ref_to_parameters->{'Geno2pfas_input_file'}) or die ('Error: Cannot create/open file: ' . $Ref_to_parameters->{'Geno2pfas_input_file'} . "\n");
	print G2PFAS_INPUT $GENO_informations;
	close (G2PFAS_INPUT);
	
	return 0;
}

##################################################
## Tools execution related functions
##################################################

sub executeGeno2pfas {
	
	# Recovers parameters
	my $Ref_to_parameters = shift;
	
	# Command line building
	my $cmd = 'geno2pfas ' . $Ref_to_parameters->{'Geno2pfas_input_file'} . ' ' . $Ref_to_parameters->{'Geno2pfas_threshold'} . ' 1> ' . $Ref_to_parameters->{'Geno2pfas_stdout_file'} . ' 2> ' . $Ref_to_parameters->{'Geno2pfas_error_file'};
		
	print '  -> Geno2pfas will be executed with the following command line: ' . $cmd . "\n" if $DEBUG;
	
	system ($cmd);
	
	return 0;
}

##################################################
## Output mangement related functions
##################################################

sub dealWithOutputStatus {
	
	# Recovers parameters
	my ($New_status, $Contig_number, $Contig_name, $Ref_to_status, $Ref_to_parameters) = @_;
	
	switch ($New_status) {
		case "Ok" {
			print '   --> Execution complete !!' . "\n\n"; 
			# Add unitary result to the global output file
			addDataToGlobalOutputFile($Ref_to_parameters->{'Geno2pfas_output_file'}, $Ref_to_parameters->{'PFAS_file'}, $Contig_number);
		}
		case "No_output" {
			print '   --> There are no Geno2pfas results for contig ' . $Contig_number . ' (' . $Contig_name . ")..\n\n";
			# Regroup contigs for which there is no normal output (for relaunch)
			addDataToGlobalOutputFile($Ref_to_parameters->{'Geno2pfas_input_file'}, $Ref_to_parameters->{'Contig_no_output'}, $Contig_number) if ($Ref_to_parameters->{'Execution_mode'} eq 'local');
		}
		case "Warning" {
			print '   --> Execution complete (with warnings) ! Results might be uncomplete ! Please check the error file for contig ' . $Contig_number . ' (' . $Contig_name . ")..\n\n";
			# Add unitary result to the global output file
			addDataToGlobalOutputFile($Ref_to_parameters->{'Geno2pfas_output_file'}, $Ref_to_parameters->{'PFAS_file'}, $Contig_number); 
			# Regroup contigs that generate warnings (for relaunch)
			addDataToGlobalOutputFile($Ref_to_parameters->{'Geno2pfas_input_file'}, $Ref_to_parameters->{'Contig_with_warning'}, $Contig_number) if ($Ref_to_parameters->{'Execution_mode'} eq 'local');
			# Add new warnings to the global error file
			addDataToGlobalOutputFile($Ref_to_parameters->{'Geno2pfas_error_file'}, $Ref_to_parameters->{'Main_error'}, $Contig_number, '## Warning(s) for contig ' . $Contig_number . ' (' . $Contig_name . ') ##'); 
		}
		case "Segmentation_fault" {
			print '   --> A segmentation fault was raised during the processing of contig ' . $Contig_number . ' (' . $Contig_name . ")..\n\n";
			# Regroup contigs that generate segmentation faults (for relaunch)
			addDataToGlobalOutputFile($Ref_to_parameters->{'Geno2pfas_input_file'}, $Ref_to_parameters->{'Contig_with_segfault'}, $Contig_number) if ($Ref_to_parameters->{'Execution_mode'} eq 'local');
			# Add new segfaults to the global error file
			addDataToGlobalOutputFile($Ref_to_parameters->{'Geno2pfas_error_file'}, $Ref_to_parameters->{'Main_error'}, $Contig_number, '## Segmentation fault for contig ' . $Contig_number . ' (' . $Contig_name . ') ##'); 
		}
		case "Error" {
			print '   --> A fatal error occured during the processing of contig ' . $Contig_number . ' (' . $Contig_name . ")..\n\n";
			# Regroup contigs that generate error messages (for relaunch)
			addDataToGlobalOutputFile($Ref_to_parameters->{'Geno2pfas_input_file'}, $Ref_to_parameters->{'Contig_with_error'}, $Contig_number) if ($Ref_to_parameters->{'Execution_mode'} eq 'local');
			# Add new errors to the global error file
			addDataToGlobalOutputFile($Ref_to_parameters->{'Geno2pfas_error_file'}, $Ref_to_parameters->{'Main_error'}, $Contig_number, '## Error(s) for contig ' . $Contig_number . ' (' . $Contig_name . ') ##'); 
		}
		else { print "Unexpected switch case !  !" }
	}
	
	$Ref_to_status->{$New_status}++;
	
	return 0;
}

sub addDataToGlobalOutputFile {
	
	# Recovers parameters
	my ($New_data, $Global_file, $Counter, $Opt_header) = @_;
	
	# Add the content of the new result file to the global result file
	open(GLOBAL, '>>' . $Global_file) or die ('Error: Cannot create/open file: ' . $Global_file . "\n");

	if ($Opt_header) { print GLOBAL $Opt_header . "\n"; }
	
	open (DATA, '<' . $New_data) or die ('Error: Cannot open/read file: ' . $New_data . "\n");	
	while (my $line = <DATA>){
		print GLOBAL $line;
	}
	close (DATA);
	
	close(GLOBAL);

	return 0;
}

##################################################
## Help/Check related functions
##################################################

sub displayHelpMessage {

	print('#################################################' . "\n");
	print('#    Geno2pfas Galaxy Wrapper - Help section    #' . "\n"); 
	print('#################################################' . "\n\n");

	print "Usage example (short names): geno2pfas_wrapper.pl -geno sample.alr_gen -sill " . $Default_threshold . " -del yes -out sample.pfas\n\n";
	
	print "Usage example (long names): geno2pfas_wrapper.pl -genofile sample.alr_gen -threshold " . $Default_threshold . " -clean yes -output sample.pfas\n\n\n";
	
	print "Here is the list of authorized parameters :\n\n";
	
	print " -help => Display this help message.\n\n";
	
	print " -genofile/-geno file => Path and name of the input file in ALR_GEN format (mandatory).\n\n";
	
	print " -threshold/-sill percentage (float) => Minimum posterior probability of a genotype required for not writing a \"N\" in the generated sequences (optional). Range of possible values [0 to 1].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_threshold . "\n\n";
	
	print " -clean/-del string => Activate/desactivate temporary files removal (optional). Possible values are [yes|no].\n";
	print "\tBy default, temporary files will not be conserved for execution instance without error.\n\n";
	
	print " -output/-out string => Path and name of the final output file in Pseudofasta format (optional).\n";
	print "\tIf not specified, a default name (derived from the name of the ALR_GEN input file) will be used.\n\n";
	
	exit (1);
}

sub checkParameters {
	
	# Recovers parameters
	my $Ref_to_parameters = shift;
	
	# Execution mode
	if (!defined($Ref_to_parameters->{'Execution_mode'}) || $Ref_to_parameters->{'Execution_mode'} eq "") {
		$Ref_to_parameters->{'Execution_mode'} = 'local';
	}
	
	# Deal with global options (Input, Output and cleaning)
	if (!defined($Ref_to_parameters->{'ALR_GEN_file'})) {
		print 'Error: No .alr_gen input file defined !' . "\n\n";
		displayHelpMessage();
	} else {
		checkInputFile($Ref_to_parameters->{'ALR_GEN_file'}, '.alr_gen');
	}
	
	if (!defined($Ref_to_parameters->{'PFAS_file'}))  {
		print 'Warning: The name of the main output file has been omited ! A default name (derived from the name of the ALR_GEN input file) will be used !' . "\n\n";	
		$Ref_to_parameters->{'PFAS_file'} = basename($Ref_to_parameters->{'ALR_GEN_file'}) .  '.pfas';
	}

	if (!defined($Ref_to_parameters->{'clean_tmp_files'}))  {
		print 'Warning: The clean/del parameter has been omited ! Temporary files will not be conserved !' . "\n\n";	
		$Ref_to_parameters->{'clean_tmp_files'} = 'yes';
		
	} elsif ($Ref_to_parameters->{'clean_tmp_files'} !~ /^(yes|no)$/i) {
		print 'Warning: The value (' . $Ref_to_parameters->{'clean_tmp_files'} . ') of the clean/del parameter given to this module is not valid ! Temporary files will not be conserved !' . "\n\n";
		$Ref_to_parameters->{'clean_tmp_files'} = 'yes';
	}


	# Deal with Geno2pfas specific options
	if (!defined($Ref_to_parameters->{'Geno2pfas_threshold'})) {
		print 'Warning: No threshold specified ! Default value ("' . $Default_threshold . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Geno2pfas_threshold'} = $Default_threshold;
		
	} elsif ($Ref_to_parameters->{'Geno2pfas_threshold'} < 0 || $Ref_to_parameters->{'Geno2pfas_threshold'} > 1) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Geno2pfas_threshold'} . ') of the Geno2pfas "threshold" parameter given to this module is not a valid numeric value between 0 and 1! Default value ("' . $Default_threshold . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Geno2pfas_threshold'} = $Default_threshold;
	}

	return 0;
}

sub checkInputFile {
	
	# Recovers parameters
	my ($file, $type) = @_;
	
	if (! -e $file || -z $file) {
		print 'Fatal error: The ' . $type . ' file specified in the command line does not exist or is empty ! Please check file name, path and content !!' . "\n";
		print 'Selected file: ' . basename($file) . "\n";
		print 'Corresponding directory: ' . dirname($file) . "\n\n";
		exit(1);
	}
	
	return 0;
}

sub checkOutputFiles {
	
	# Recovers parameters
	my ($Output_file, $Error_file) = @_;
	
	# Initializations
	my $Status = 'Unknown';

	if (-e $Error_file && ! -z $Error_file) {
		
		# Check for critical errors
		open (ERROR_FILE, '<' . $Error_file) or die ('Error: Cannot open/read file: ' . $Error_file . "\n");
		while (my $error_line = <ERROR_FILE>){
			if ($error_line =~ /Segmentation[\s_]*fault|segfault|Erreur de segmentation/i) { $Status = 'Segmentation_fault'; }
		}				
		close (ERROR_FILE);
		
		if ($Status ne 'Segmentation_fault') {
			
			if ( -e $Output_file && ! -z $Output_file) {
				$Status = 'Warning';
			} else {
				$Status = 'Error';
			}
		}
		
	} else {
		
		if ( -e $Output_file && ! -z $Output_file) {
			$Status = 'Ok';
		} else {
			$Status = 'No_output';
		}
	}
	
	return $Status;
}

##################################################
## Basic common functions
##################################################

sub humanReadableDate {
	
	# Recovers parameters
	my $time = shift || time;
	
	# Split time string
	my ($seconde, $minute, $heure, $jour, $mois, $annee, $jour_semaine, $jour_annee, $heure_hiver_ou_ete) = localtime($time);
	$mois  += 1;
	$annee += 1900;

	# On rajoute 0 si le chiffre est compris entre 1 et 9
	foreach ( $seconde, $minute, $heure, $jour, $mois, $annee ) { s/^(\d)$/0$1/; }
	
	return "$jour-$mois-$annee" . '_' . "$heure:$minute:$seconde";
}

sub cleanItAfterMe {
	
	# Remove all files given as arguments to this function
	foreach my $File_to_delete (@_) {
		if (-e  $File_to_delete) {
			unlink($File_to_delete);
		}
	}
	
	return 0;
}
