#!/usr/bin/perl

## Date and Version
my $version= '1.2';
my $version_date = '17_fev_2012';

##################################################
## Documentation POD
##################################################

##################################################
## Modules
##################################################
## Perl modules
use strict;
use warnings;
use diagnostics;

use Getopt::Long;
use File::Basename;
use Switch;
use Data::Dumper;

# BioPerl modules
use Bio::SeqIO;
use Bio::DB::Fasta;

##################################################
## Globals variables
##################################################

# Initializations
my $Start_time_hr = humanReadableDate();
my $help = undef();
my $DEBUG = 0;
my %Parameters;
my %findmode_text = ( '0' => 'Translation of regions between STOP codons', '1' => 'Translation of regions between START and STOP codons', '2' => 'Nucleic sequences between STOP codons', '3' => 'Nucleic sequences between START and STOP codons', '4' => 'Nucleotides flanking START codons', '5' => 'Nucleotides flanking initial STOP codons', '6' => 'Nucleotides flanking ending STOP codons' );

# Default values
my $Default_Genetic_code = 0;
my $Default_ORF_minsize = 30;
my $Default_ORF_maxsize = 1000000;
my $Default_Output_type = 3;
my $Default_Flanking = 100;
my $Default_Start_with_metionine = 'yes';
my $Default_Circular_sequence = 'no';
my $Default_Search_reverse = 'yes';
my $Default_ORF_to_keep = 'all';

##################################################
## Manage options
##################################################

# Get options from command line
GetOptions (	'help|h'					=> \$help,
			'sequence|fasta=s'		=> \$Parameters{'Fasta_file'},
			'genetic_code|code:i'		=> \$Parameters{'Genetic_code'},
			'minsize|min:i'			=> \$Parameters{'ORF_minsize'},
			'maxsize|max:i'			=> \$Parameters{'ORF_maxsize'},
			'output_type|find:i'		=> \$Parameters{'Output_type'},
			'flanking:i'				=> \$Parameters{'Flanking'},
			'methionine|meth:s'		=> \$Parameters{'Start_with_metionine'},
			'circular:s'				=> \$Parameters{'Circular_sequence'},
			'reverse:s'				=> \$Parameters{'Search_reverse'},
			'keep:s'					=> \$Parameters{'ORF_to_keep'},
			'output|out:s'			=> \$Parameters{'Main_output'}
		);

# Display help message if needed
if (defined($help)) {
	displayHelpMessage();
}

# Check parameters
checkParameters(\%Parameters);

##################################################
## Main code
##################################################

# Welcome message
print "############################################################\n";
print "#      Welcome in getORF Galaxy Wrapper (Version $version)      #\n"; 
print "############################################################\n";
print "Start time: " . $Start_time_hr . "\n\n";

# Verbose
print 'Selected Fasta input file: ' . $Parameters{'Fasta_file'} . "\n";
print 'Selected output type: ' . $findmode_text{$Parameters{'Output_type'}} . "\n\n";

# Execute getORF
print ' ORF search procedure' . "\n";
execute_getORF(\%Parameters);

# Check output
my $Output_status = checkOutputFiles($Parameters{'getORF_output_file'}, $Parameters{'getORF_error_file'});

switch ($Output_status) {
	case "Ok" { print '   --> Execution complete !!' . "\n\n"; }
	case "No_output" { print '   --> There are no getORF results..' . "\n\n"; }
	case "Warning" { print '   --> Execution complete (with warnings) ! Results might be uncomplete..' . "\n\n"; }
	case "Segmentation_fault" {	print '   --> A segmentation fault was raised during the execution process..' . "\n\n"; }
	case "Error" { print '   --> A fatal error occured during the execution process..' . "\n\n"; }
	else { print "Unexpected switch case !"; }
}

# Get total number of ORFs discovered by getORF
my $Total_NB_ORF = getTotalNumberOfORF(\%Parameters);

# Keep the best ORF for each contig/sequence if needed
if ($Total_NB_ORF > 0) {
	
	if ($Parameters{'ORF_to_keep'} eq "best") {
		# Parse getORF brut output file and extract the longuest ORFs
		my $BestORF_ref = getBestORFs(\%Parameters, $Total_NB_ORF);
		writeBestORFs($BestORF_ref, \%Parameters);
		
	} else {
		# Display the number of discovered ORF
		print ' ' . $Total_NB_ORF . ' ORF(s) discovered' . "\n\n";
		
		# No filtering needed
		rename($Parameters{'getORF_output_file'}, $Parameters{'Main_output'});
	}
	
} else {
	print STDERR 'Error: No ORF discovered by getORF !!' . "\n\n"
}

print 'Global output file: ' . $Parameters{'Main_output'} . "\n\n";

print "\nStop time: " . humanReadableDate() . "\n";
print "############################################################\n";
print "#                      End of execution                    #\n"; 
print "############################################################\n";

##################################################
## ORF related functions
##################################################

sub execute_getORF {
	
	# Recovers parameters
	my ($Ref_to_parameters) = shift;
	
	# Initializations
	$Ref_to_parameters->{'getORF_input_file_basename'} = basename($Ref_to_parameters->{'Fasta_file'});
	$Ref_to_parameters->{'getORF_stdout_file'} = $Ref_to_parameters->{'getORF_input_file_basename'} . '.out';
	$Ref_to_parameters->{'getORF_error_file'} = $Ref_to_parameters->{'getORF_input_file_basename'} . '.err';
	$Ref_to_parameters->{'getORF_output_file'} = $Ref_to_parameters->{'getORF_input_file_basename'} . '.orf';
	
	# Command line building
	my $cmd = 'getorf' . ' -sequence ' . $Ref_to_parameters->{'Fasta_file'};
	$cmd .= ' -outseq ' .$Ref_to_parameters->{'getORF_output_file'};
	$cmd .= ' -table ' . $Ref_to_parameters->{'Genetic_code'};
	$cmd .= ' -minsize ' . $Ref_to_parameters->{'ORF_minsize'};
	$cmd .= ' -maxsize ' . $Ref_to_parameters->{'ORF_maxsize'};
	$cmd .= ' -find ' . $Ref_to_parameters->{'Output_type'};	
	if ($Ref_to_parameters->{'Output_type'} == 4 || $Ref_to_parameters->{'Output_type'} == 5 || $Ref_to_parameters->{'Output_type'} == 6) { $cmd .= ' -flanking ' . $Ref_to_parameters->{'Flanking'}; }	
	if ($Ref_to_parameters->{'Start_with_metionine'} eq "yes") { $cmd .= ' -methionine'; } else { $cmd .= ' -nomethionine'; }	
	if ($Ref_to_parameters->{'Search_reverse'} eq "yes") { $cmd .= ' -reverse'; } else { $cmd .= ' -noreverse'; }	
	if ($Ref_to_parameters->{'Circular_sequence'} eq "yes") { $cmd .= ' -circular'; }	
	$cmd .= ' 1> ' . $Ref_to_parameters->{'getORF_stdout_file'} . ' 2> ' . $Ref_to_parameters->{'getORF_error_file'};

	print '  -> getORF will be executed with the following command line: ' . $cmd . "\n" if $DEBUG;

	system ($cmd);
	
	return 0;
}

sub getBestORFs {
	
	# Recovers parameters
	my ($Ref_to_parameters, $Nb_Raw_ORF) = @_;
	
	# Initializations
	my %Best_ORFs;
	
	print ' Search for the best/longest ORF of each contig' . "\n";
	
	# Browse the ORF file and collect the start and stop position of the longest ORF for each contig
	my $SeqIO_input_object  = Bio::SeqIO->new ( '-file' => $Ref_to_parameters->{'getORF_output_file'}, '-format' => 'FASTA' );
	
	while (my $Current_ORF = $SeqIO_input_object->next_seq) {

		# Extract the contig name from ORF name
		my ($Contig_name) = ($Current_ORF->display_id() =~ /(.*)_\d+$/);
		my $ORF_length = $Current_ORF->length();
		
		# If we are on the first ORF of a contig we store its information
		if (! defined($Best_ORFs{$Contig_name})) {
			$Best_ORFs{$Contig_name} = { 'ORF_desc' => $Current_ORF->description(), 'ORF_seq' => $Current_ORF->seq(), 'ORF_length' => $ORF_length };
		
		# Otherwise we check if the current ORF is the longer than the previous one
		# If that's the case we considered it as the new best ORF
		} else {
			if ($Best_ORFs{$Contig_name}->{'ORF_length'} < $ORF_length) {
				$Best_ORFs{$Contig_name} = { 'ORF_desc' => $Current_ORF->description(), 'ORF_seq' => $Current_ORF->seq(), 'ORF_length' => $ORF_length };
			}
		}
	}

	my $Number_of_Best_ORF = scalar(keys %Best_ORFs);
	
	if ($Number_of_Best_ORF > 0) {
		print '   --> Research complete: ' . $Number_of_Best_ORF . ' best ORF(s) collected (from ' . $Nb_Raw_ORF . ' raw ORF(s))' . "\n\n";
	} else {
		print '   --> No best ORF found..' . "\n\n";
	}
	
	return \%Best_ORFs;
}

sub writeBestORFs {

	# Recovers parameters
	my ($Ref_to_BestORFs, $Ref_to_parameters) = @_;
	
	# Initializations
	my $sort_rule = 'InAlphaNumericOrder';
	
	# Verbose
	print ' Write the best/longest ORF of each contig into the main output file' . "\n";

	# Create a Bio::SeqIO object for output (Fasta)
	my $SeqIO_output_object = Bio::SeqIO->new( -format => 'FASTA', -file => '>' . $Ref_to_parameters->{'Main_output'});
	
	# Examinate one key of the bestORF hash table to determine which sort rules is needed
	my $Tests_keys = (keys(%{$Ref_to_BestORFs}))[0];
	
	if ($Tests_keys =~ /Contig.*/) { $sort_rule = 'ByContigNumberAsc'; }
	elsif ($Tests_keys =~ /gi.*/) { $sort_rule = 'ByGenbankIdentifierAsc'; }
	
	# Go through the list of Best ORFs and write them in the output file	
	foreach my $Best_ORF (sort $sort_rule (keys %{$Ref_to_BestORFs})) {
		my $Sequence_object = Bio::PrimarySeq->new ( '-display_id' => $Best_ORF , '-seq' => $Ref_to_BestORFs->{$Best_ORF}->{'ORF_seq'}, '-description' => $Ref_to_BestORFs->{$Best_ORF}->{'ORF_desc'} . '(Length:' . $Ref_to_BestORFs->{$Best_ORF}->{'ORF_length'} . ')' );
		$SeqIO_output_object->write_seq($Sequence_object);
	}
	
	print '   --> Done !' . "\n\n";
	
	return 0;
}

# TODO: Use a pure Perl/BioPerl method to get the total number of ORF discovered by getORF
sub getTotalNumberOfORF {
	
	# Recovers parameters
	my $Ref_to_parameters = shift;
	
	# Get number of ORF with grep and wc
	my $Raw_ORF_Counter_with_grep = `grep '>' $Ref_to_parameters->{'getORF_output_file'} | wc -l`;
	
	chomp($Raw_ORF_Counter_with_grep);
	
	return $Raw_ORF_Counter_with_grep;
}

##################################################
## Custom sort rules
##################################################

sub InAlphaNumericOrder {
	$a cmp $b;
}

# sort using explicit subroutine name
sub ByContigNumberAsc {
	($a =~ /Contig(\d+)$/)[0] <=> ($b =~ /Contig(\d+)$/)[0];
}

sub ByGenbankIdentifierAsc {
	($a =~ /gi.(\d+)/)[0] <=> ($b =~ /gi.(\d+)/)[0];
}


##################################################
## Help/Check related functions
##################################################

sub displayHelpMessage {

	print('##################################################' . "\n");
	print('#      getORF Galaxy Wrapper - Help section      #' . "\n"); 
	print('##################################################' . "\n\n");

	print "Usage example (short names): getorf_wrapper.pl -fasta my_fasta_file -code " . $Default_Genetic_code . " -min " . $Default_ORF_minsize . " -max " . $Default_ORF_maxsize . " -find " . $Default_Output_type . " -meth " . $Default_Start_with_metionine . " -circular " . $Default_Circular_sequence . " -reverse " . $Default_Search_reverse . " -keep " . $Default_ORF_to_keep . " -out getORF_output.orf " . "\n\n";
	
	print "Usage example (long names): getorf_wrapper.pl -sequence my_fasta_file -genetic_code " . $Default_Genetic_code . " -minsize " . $Default_ORF_minsize . " -maxsize " . $Default_ORF_maxsize . " -output_type " . $Default_Output_type . " -methionine_start " . $Default_Start_with_metionine . " -circular " . $Default_Circular_sequence . " -reverse " . $Default_Search_reverse . " -keep " . $Default_ORF_to_keep . " -out getORF_output.orf " . "\n\n\n";
	
	print "Here is the list of authorized parameters :\n\n";
	
	print " -help => Display this help message.\n\n";
	
	print " -sequence/-fasta file => Path and name of the input Fasta file (mandatory).\n\n";
	
	print " -genetic_code/-code integer => Genetic code to use (optional). Possible values are [0-6|9-16|21-23] - .\n";
	print "     If set to 0: Standard code.\n";
	print "     Please, try 'getorf -help' for more details.\n";
	print "\tIf not specified, the following default value will be used: " . $Default_Genetic_code . "\n\n";

	print " -minsize/-min integer => Minimum nucleotide size of ORF to report (optional). Possible values are [All positive integers].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_ORF_minsize . "\n\n";

	print " -maxsize/-max integer => Maximum nucleotide size of ORF to report (optional). Possible values are [All positive integers].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_ORF_maxsize . "\n\n";

	print " -output_type/-find integer => Define the type of ORF that getORF must find and output (getORF find option) (optional). Possible values are [0-6].\n";
	print "     If set to 0: Translation of regions between STOP codons.\n";
	print "     If set to 1: Translation of regions between START and STOP codons.\n";
	print "     If set to 2: Nucleic sequences between STOP codons.\n";
	print "     If set to 3: Nucleic sequences between START and STOP codons.\n";
	print "     If set to 4: Nucleotides flanking START codons.\n";
	print "     If set to 5: Nucleotides flanking initial STOP codons.\n";
	print "     If set to 6: Nucleotides flanking ending STOP codons.\n";
	print "\tIf not specified, the following default value will be used: " . $Default_Output_type . "\n\n";

	print " -flanking integer => Size of the flanking region(s) to collect (Only used if -output_type/-type option is set to 4,5,6) (optional). Possible values are [All positive integers].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_Flanking . "\n\n";

	print " -methionine/-meth string => Define if the ORF must start by a methionine (yes), or not (no) (optional). Possible values are [yes|no].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_Start_with_metionine . "\n\n";

	print " -circular string => Define if the sequence is circular (yes), or not (no) (optional). Possible values are [yes|no].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_Circular_sequence . "\n\n";

	print " -reverse string => Define if the ORFs must also be searched on the reverse complement sequence (yes), or not (no) (optional). Possible values are [yes|no].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_Search_reverse . "\n\n";

	print " -keep string => Define if you want to keep all ORF for each sequence/contig, or only the best (optional). Possible values are [all|best].\n";
	print "\tIf not specified, the following default value will be used: " . $Default_ORF_to_keep . "\n\n";
	
	print " -output/-out string => Path and name of the final output file in Fasta format (optional).\n";
	print "\tIf not specified, a default name (derived from the name of the Fasta input file) will be used.\n\n";

	exit (1);
}

sub checkParameters {
	
	# Recovers parameters
	my $Ref_to_parameters = shift;

	# Deal with global options (Input, Output and cleaning)
	if (!defined($Ref_to_parameters->{'Fasta_file'})) {
		print 'Error: No Fasta input file defined !' . "\n\n";
		displayHelpMessage();
	} else {
		checkInputFile($Ref_to_parameters->{'Fasta_file'}, 'Fasta');
	}
	
	if (!defined($Ref_to_parameters->{'Main_output'}))  {
		print 'Warning: The name of the main output file has been omited ! A default name (derived from the name of the Fasta input file) will be used !' . "\n\n";	
		$Ref_to_parameters->{'Main_output'} = basename($Ref_to_parameters->{'Fasta_file'}) .  '.orf';
	}
	
	# Deal with getORF specific options	
	if (!defined($Ref_to_parameters->{'Genetic_code'}))  {
		print 'Warning: The "genetic_code|code" parameter has been omited ! Default value ("' . $Default_Genetic_code . '") will be used...' . "\n\n";
		$Ref_to_parameters->{'Genetic_code'} = $Default_Genetic_code;
		
	} elsif ($Ref_to_parameters->{'Genetic_code'} < 0 || ($Ref_to_parameters->{'Genetic_code'} > 6 && $Ref_to_parameters->{'Genetic_code'} < 9) || ($Ref_to_parameters->{'Genetic_code'} >16 && $Ref_to_parameters->{'Genetic_code'} < 21) || $Ref_to_parameters->{'Genetic_code'} > 23) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Genetic_code'} . ') of the "genetic_code|code" parameter given to this module is not valid ! Default value ("' . $Default_Genetic_code . '") will be used...' . "\n\n";
		$Ref_to_parameters->{'Genetic_code'} = $Default_Genetic_code;
	}

	if (!defined($Ref_to_parameters->{'ORF_minsize'})) {
		print 'Warning: No "minsize|min" parameter specified ! Default value ("' . $Default_ORF_minsize . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'ORF_minsize'} = $Default_ORF_minsize;
		
	} elsif ($Ref_to_parameters->{'ORF_minsize'} !~ /^[0-9]+$/ || $Ref_to_parameters->{'ORF_minsize'} <= 0) {
		print 'Warning: The value (' . $Ref_to_parameters->{'ORF_minsize'} . ') of the "minsize|min" parameter given to this module is not valid ! Default value ("' . $Default_ORF_minsize . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'ORF_minsize'} = $Default_ORF_minsize;
	}
	
	if (!defined($Ref_to_parameters->{'ORF_maxsize'})) {
		print 'Warning: No "maxsize|max" parameter specified ! Default value ("' . $Default_ORF_maxsize . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'ORF_maxsize'} = $Default_ORF_maxsize;
		
	} elsif ($Ref_to_parameters->{'ORF_maxsize'} !~ /^[0-9]+$/ || $Ref_to_parameters->{'ORF_maxsize'} <= 0) {
		print 'Warning: The value (' . $Ref_to_parameters->{'ORF_maxsize'} . ') of the "maxsize|max" parameter given to this module is not valid ! Default value ("' . $Default_ORF_maxsize . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'ORF_maxsize'} = $Default_ORF_maxsize;
	}
	
	if ($Ref_to_parameters->{'ORF_minsize'} > $Ref_to_parameters->{'ORF_maxsize'}) {
		print 'Error: ORF minsize cannot be greater than maxsize ! Default minsize ("' . $Default_ORF_minsize . '") and default maxsize ("' . $Default_ORF_maxsize . '") will be used...' . "\n\n";
	}

	if (!defined($Ref_to_parameters->{'Output_type'})) {
		print 'Warning: No "output_type|find" parameter specified ! Default value ("' . $Default_Output_type . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Output_type'} = $Default_Output_type;
		
	} elsif ($Ref_to_parameters->{'Output_type'} < 0 || $Ref_to_parameters->{'Output_type'} > 6) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Output_type'} . ') of the "output_type|find" parameter given to this module is not valid ! Default value ("' . $Default_Output_type . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Output_type'} = $Default_Output_type;
		
	} elsif ($Ref_to_parameters->{'Output_type'} == 4 || $Ref_to_parameters->{'Output_type'} == 5 || $Ref_to_parameters->{'Output_type'} == 6) {
		
		if (!defined($Ref_to_parameters->{'Flanking'})) {
			print 'Warning: No "flanking" parameter specified ! Default value ("' . $Default_Flanking . '") will be used...' . "\n\n";	
			$Ref_to_parameters->{'Flanking'} = $Default_Flanking;
			
		} elsif ($Ref_to_parameters->{'Flanking'} !~ /^[0-9]+$/ || $Ref_to_parameters->{'Flanking'} <= 0) {
			print 'Warning: The value (' . $Ref_to_parameters->{'Flanking'} . ') of the "flanking" parameter given to this module is not valid ! Default value ("' . $Default_Flanking . '") will be used !' . "\n\n";
			$Ref_to_parameters->{'Flanking'} = $Default_Flanking;
		}
	}	
	
	if (!defined($Ref_to_parameters->{'Start_with_metionine'})) {
		print 'Warning: No "methionine|meth" parameter specified ! Default value ("' . $Default_Start_with_metionine . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Start_with_metionine'} = $Default_Start_with_metionine;
		
	} elsif ($Ref_to_parameters->{'Start_with_metionine'} !~ /^(yes|no)$/i) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Start_with_metionine'} . ') of the "methionine|meth" parameter given to this module is not valid ! Default value ("' . $Default_Start_with_metionine . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Start_with_metionine'} = $Default_Start_with_metionine;
	}
	
	if (!defined($Ref_to_parameters->{'Circular_sequence'})) {
		print 'Warning: No "circular" parameter specified ! Default value ("' . $Default_Circular_sequence . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Circular_sequence'} = $Default_Circular_sequence;
		
	} elsif ($Ref_to_parameters->{'Circular_sequence'} !~ /^(yes|no)$/i) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Circular_sequence'} . ') of the "circular" parameter given to this module is not valid ! Default value ("' . $Default_Circular_sequence . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Circular_sequence'} = $Default_Circular_sequence;
	}
	
	if (!defined($Ref_to_parameters->{'Search_reverse'})) {
		print 'Warning: No "reverse" parameter specified ! Default value ("' . $Default_Search_reverse . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'Search_reverse'} = $Default_Search_reverse;
		
	} elsif ($Ref_to_parameters->{'Search_reverse'} !~ /^(yes|no)$/i) {
		print 'Warning: The value (' . $Ref_to_parameters->{'Search_reverse'} . ') of the "reverse" parameter given to this module is not valid ! Default value ("' . $Default_Search_reverse . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'Search_reverse'} = $Default_Search_reverse;
	}
	
	if (!defined($Ref_to_parameters->{'ORF_to_keep'})) {
		print 'Warning: No "keep" parameter specified ! Default value ("' . $Default_ORF_to_keep . '") will be used...' . "\n\n";	
		$Ref_to_parameters->{'ORF_to_keep'} = $Default_ORF_to_keep;
		
	} elsif ($Ref_to_parameters->{'ORF_to_keep'} !~ /^(all|best)$/i) {
		print 'Warning: The value (' . $Ref_to_parameters->{'ORF_to_keep'} . ') of the "keep" parameter given to this module is not valid ! Default value ("' . $Default_ORF_to_keep . '") will be used !' . "\n\n";
		$Ref_to_parameters->{'ORF_to_keep'} = $Default_ORF_to_keep;
	}
}

sub checkInputFile {
	
	# Recovers parameters
	my ($file, $type) = @_;
	
	if (! -e $file || -z $file) {
		print 'Fatal error: The ' . $type . ' file specified in the command line does not exist or is empty ! Please check file name, path and content !!' . "\n";
		print 'Selected file: ' . basename($file) . "\n";
		print 'Corresponding directory: ' . dirname($file) . "\n\n";
		exit(1);
	}
	
	return 0;
}

sub checkOutputFiles {
	
	# Recovers parameters
	my ($Output_file, $Error_file) = @_;
	
	# Initializations
	my ($Status, $bad_output_channel) = ('Unknown', 'no');
	my $NB_error_line = 0;	

	if (-e $Error_file && ! -z $Error_file) {
		
		# Check for critical errors
		open (ERROR_FILE, '<' . $Error_file) or die ('Error: Cannot open/read file: ' . $Error_file . "\n");
		while (my $error_line = <ERROR_FILE>){
			if ($error_line =~ /Segmentation[\s_]*fault|segfault|Erreur de segmentation/i) { $Status = 'Segmentation_fault'; }
			if ($error_line =~ /Finds and extracts open reading frames/i) { $bad_output_channel = 'yes'; }
			$NB_error_line++;
		}				
		close (ERROR_FILE);
		
		if ($Status ne 'Segmentation_fault') {
			
			if ( -e $Output_file && ! -z $Output_file) {
				if ($bad_output_channel eq 'yes' && $NB_error_line == 1) {
					$Status = 'Ok';
				} else {
					$Status = 'Warning';
				}
			} else {
				$Status = 'Error';
			}
		}
		
	} else {
		
		if ( -e $Output_file && ! -z $Output_file) {
			$Status = 'Ok';
		} else {
			$Status = 'No_output';
		}
	}
	
	return $Status;
}

##################################################
## Basic common functions
##################################################

sub humanReadableDate {
	
	# Recovers parameters
	my $time = shift || time;
	
	# Split time string
	my ($seconde, $minute, $heure, $jour, $mois, $annee, $jour_semaine, $jour_annee, $heure_hiver_ou_ete) = localtime($time);
	$mois  += 1;
	$annee += 1900;

	# On rajoute 0 si le chiffre est compris entre 1 et 9
	foreach ( $seconde, $minute, $heure, $jour, $mois, $annee ) { s/^(\d)$/0$1/; }
	
	return "$jour-$mois-$annee" . '_' . "$heure:$minute:$seconde";
}
